package lcsb.mapviewer.converter.model.sbml;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.awt.Color;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.junit.Test;
import org.reflections.Reflections;
import org.sbml.jsbml.SBMLDocument;
import org.sbml.jsbml.ext.multi.MultiModelPlugin;
import org.sbml.jsbml.ext.multi.MultiSpeciesPlugin;
import org.sbml.jsbml.ext.multi.MultiSpeciesType;
import org.sbml.jsbml.ext.multi.PossibleSpeciesFeatureValue;
import org.sbml.jsbml.ext.multi.SpeciesFeatureType;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.common.comparator.LineComparator;
import lcsb.mapviewer.common.comparator.ListComparator;
import lcsb.mapviewer.converter.ConverterParams;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.model.graphics.PolylineData;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelComparator;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.reaction.AbstractNode;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.ReactionNode;
import lcsb.mapviewer.model.map.reaction.type.StateTransitionReaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.map.species.field.ModificationState;
import lcsb.mapviewer.model.map.species.field.PositionToCompartment;
import lcsb.mapviewer.model.map.species.field.Residue;

public class SbmlExporterTest extends SbmlTestFunctions {
  Logger logger = Logger.getLogger(SbmlExporterTest.class);
  SbmlParser parser = new SbmlParser();
  SbmlExporter exporter;

  ModelComparator comparator = new ModelComparator();

  public SbmlExporterTest() {
    exporter = new SbmlExporter();
  }

  @Test
  public void testExportCompartment() throws Exception {
    Model model = getModelAfterSerializing("testFiles/layoutExample/CompartmentGlyph_Example_level2_level3.xml");
    assertNotNull(model);
    assertEquals(1, model.getCompartments().size());
    Compartment compartment = model.getElementByElementId("CompartmentGlyph_1");
    assertNotNull(compartment);
    assertNotNull(compartment.getX());
    assertNotNull(compartment.getY());
    assertTrue(compartment.getWidth() > 0);
    assertTrue(compartment.getHeight() > 0);
    assertNotNull(model.getHeight());
    assertNotNull(model.getWidth());
    assertTrue(model.getWidth() >= compartment.getX() + compartment.getWidth());
    assertTrue(model.getHeight() >= compartment.getY() + compartment.getHeight());
    assertFalse(compartment.getClass().equals(Compartment.class));
  }

  private Model getModelAfterSerializing(String filename) throws Exception {
    Model originalModel = parser.createModel(new ConverterParams().filename(filename));
    return getModelAfterSerializing(originalModel);
  }

  private Model getModelAfterSerializing(Model originalModel) throws Exception {
    exporter.setProvideDefaults(false);
    String xml = exporter.toXml(originalModel);
    ByteArrayInputStream stream = new ByteArrayInputStream(xml.getBytes("UTF-8"));
    Model result = parser.createModel(new ConverterParams().inputStream(stream));
    // showImage(originalModel);
    // showImage(result);
    return result;
  }

  @Test
  public void testExportSpecies() throws Exception {
    Model model = getModelAfterSerializing("testFiles/layoutExample/SpeciesGlyph_Example_level2_level3.xml");
    assertNotNull(model);
    assertEquals(1, model.getElements().size());
    Species glucoseSpecies = model.getElementByElementId("SpeciesGlyph_Glucose");
    assertNotNull(glucoseSpecies.getX());
    assertNotNull(glucoseSpecies.getY());
    assertTrue(glucoseSpecies.getWidth() > 0);
    assertTrue(glucoseSpecies.getHeight() > 0);
    assertNotNull(model.getHeight());
    assertNotNull(model.getWidth());
    assertTrue(model.getWidth() >= glucoseSpecies.getX() + glucoseSpecies.getWidth());
    assertTrue(model.getHeight() >= glucoseSpecies.getY() + glucoseSpecies.getHeight());
    assertFalse(glucoseSpecies.getClass().equals(Species.class));
  }

  @Test
  public void testExportSpeciesInCompartments() throws Exception {
    Model model = getModelAfterSerializing("testFiles/layoutExample/SpeciesGlyph_Example.xml");
    assertNotNull(model);
    assertEquals(2, model.getElements().size());
    Compartment compartment = model.getElementByElementId("Yeast");
    assertNotNull(compartment.getX());
    assertNotNull(compartment.getY());

    assertNotNull(compartment.getX() >= 0);
    assertNotNull(compartment.getY() >= 0);
    assertTrue(compartment.getWidth() > 0);
    assertTrue(compartment.getHeight() > 0);
    assertNotNull(model.getHeight());
    assertNotNull(model.getWidth());
    assertTrue(model.getWidth() >= compartment.getX() + compartment.getWidth());
    assertTrue(model.getHeight() >= compartment.getY() + compartment.getHeight());
    assertFalse(compartment.getClass().equals(Compartment.class));
    assertTrue(compartment.getElements().size() > 0);
  }

  @Test
  public void testExportReaction() throws Exception {
    Model model = getModelAfterSerializing("testFiles/layoutExample/Complete_Example.xml");
    assertNotNull(model);
    assertEquals(1, model.getReactions().size());
    Reaction reaction = model.getReactions().iterator().next();
    for (ReactionNode node : reaction.getReactionNodes()) {
      assertNotNull(node.getLine());
      assertTrue(node.getLine().length() > 0);
    }
    assertEquals(2, reaction.getOperators().size());
  }

  @Test
  public void testExportModelId() throws Exception {
    Model model = new ModelFullIndexed(null);
    model.setIdModel("Test123");
    Model deserializedModel = getModelAfterSerializing(model);

    assertEquals(model.getIdModel(), deserializedModel.getIdModel());
  }

  @Test
  public void testExportModelDimension() throws Exception {
    Model model = new ModelFullIndexed(null);
    model.setWidth(200);
    model.setHeight(300);
    Model deserializedModel = getModelAfterSerializing(model);

    assertEquals(model.getWidth(), deserializedModel.getWidth());
    assertEquals(model.getHeight(), deserializedModel.getHeight());
  }

  @Test
  public void testExportEmptyModel() throws Exception {
    Model model = new ModelFullIndexed(null);
    Model deserializedModel = getModelAfterSerializing(model);

    assertNotNull(deserializedModel);
  }

  @Test
  public void testExportReactionWithLayout() throws Exception {
    Model model = createModelWithReaction();
    Reaction reaction = model.getReactions().iterator().next();

    Model deserializedModel = getModelAfterSerializing(model);
    Reaction deserializedReaction = deserializedModel.getReactions().iterator().next();

    assertEquals(reaction.getReactants().get(0).getLine().length(),
        deserializedReaction.getReactants().get(0).getLine().length(), Configuration.EPSILON);
    assertEquals(reaction.getProducts().get(0).getLine().length(),
        deserializedReaction.getProducts().get(0).getLine().length(), Configuration.EPSILON);

  }

  private Model createModelWithReaction() {
    Model model = createEmptyModel();
    GenericProtein p1 = new GenericProtein("s1");
    p1.setWidth(100);
    p1.setHeight(20);
    p1.setX(10);
    p1.setY(10);
    model.addElement(p1);
    GenericProtein p2 = new GenericProtein("s2");
    p2.setWidth(100);
    p2.setHeight(20);
    p2.setX(10);
    p2.setY(50);
    model.addElement(p2);
    Reaction reaction = new StateTransitionReaction();
    Reactant reactant = new Reactant(p1);
    Product product = new Product(p2);
    PolylineData reactantLine = new PolylineData();
    reactantLine.addPoint(new Point2D.Double(50, 20));
    reactantLine.addPoint(new Point2D.Double(50, 30));
    reactant.setLine(reactantLine);
    PolylineData productLine = new PolylineData();
    productLine.addPoint(new Point2D.Double(50, 40));
    productLine.addPoint(new Point2D.Double(50, 60));
    product.setLine(productLine);
    reaction.addReactant(reactant);
    reaction.addProduct(product);
    model.addReaction(reaction);
    return model;
  }

  @Test
  public void testExportReactionWithNotes() throws Exception {
    Model model = createModelWithReaction();
    Reaction reaction = model.getReactions().iterator().next();
    reaction.setNotes("XYZ");

    Model deserializedModel = getModelAfterSerializing(model);
    Reaction deserializedReaction = deserializedModel.getReactions().iterator().next();

    assertEquals(reaction.getNotes(), deserializedReaction.getNotes());

  }

  @Test
  public void testExportModelWithReaction() throws Exception {
    String tempFilename = File.createTempFile("tmp", ".xml").getAbsolutePath();
    SbmlParser converter = new SbmlParser();
    converter.setProvideDefaults(false);

    Model model = converter.createModel(new ConverterParams().filename("testFiles/layoutExample/example1.xml"));
    model.setName(null);

    converter.exportModelToFile(model, tempFilename);

    Model model2 = converter.createModel(new ConverterParams().filename(tempFilename));
    model2.setName(null);

    assertNotNull(model2);
    ModelComparator comparator = new ModelComparator(1.0);
    assertEquals(0, comparator.compare(model, model2));

  }

  @Test
  public void testExportModelName() throws Exception {
    String tempFilename = File.createTempFile("tmp", ".xml").getAbsolutePath();
    SbmlParser converter = new SbmlParser();
    converter.setProvideDefaults(false);

    Model model = createEmptyModel();

    converter.exportModelToFile(model, tempFilename);

    Model model2 = converter.createModel(new ConverterParams().filename(tempFilename));

    assertNotNull(model2);
    ModelComparator comparator = new ModelComparator(1.0);
    assertEquals(0, comparator.compare(model, model2));

  }

  private Model createEmptyModel() {
    Model model = new ModelFullIndexed(null);
    model.setWidth(1000);
    model.setHeight(1000);
    model.setName("UNKNOWN DISEASE MAP");
    model.setIdModel("id1");
    return model;
  }

  @Test
  public void testExportProblematicNotes() throws Exception {
    Model model = createModelWithReaction();
    Reaction reaction = model.getReactions().iterator().next();
    reaction.setNotes("X=Y<Z");

    Model deserializedModel = getModelAfterSerializing(model);
    Reaction deserializedReaction = deserializedModel.getReactions().iterator().next();

    assertEquals(reaction.getNotes(), deserializedReaction.getNotes());

  }

  @Test
  public void testColorParsing() throws Exception {
    Model model = parser.createModel(new ConverterParams().filename("testFiles/layoutExample/example1.xml"));
    for (Element element : model.getElements()) {
      element.setColor(Color.BLUE);
    }
    Model model2 = getModelAfterSerializing(model);
    assertEquals(0, comparator.compare(model, model2));
  }

  @Test
  public void testReactionColorParsing() throws Exception {
    Model model = parser.createModel(new ConverterParams().filename("testFiles/layoutExample/example1.xml"));
    for (Reaction element : model.getReactions()) {
      for (AbstractNode node : element.getNodes()) {
        node.getLine().setColor(Color.BLUE);
      }
    }
    Model model2 = getModelAfterSerializing(model);
    assertEquals(0, comparator.compare(model, model2));
  }

  @Test
  public void testReactionLineWidthParsing() throws Exception {
    Model model = parser.createModel(new ConverterParams().filename("testFiles/layoutExample/example1.xml"));
    for (Reaction element : model.getReactions()) {
      for (AbstractNode node : element.getNodes()) {
        node.getLine().setWidth(12.7);
      }
    }
    Model model2 = getModelAfterSerializing(model);
    assertEquals(0, comparator.compare(model, model2));
  }

  @Test
  public void testExportSpeciesType() throws Exception {
    Reflections reflections = new Reflections("lcsb.mapviewer.model.map");
    Set<Class<? extends Species>> classes = reflections.getSubTypesOf(Species.class);
    for (Class<? extends Element> class1 : classes) {
      if (!Modifier.isAbstract(class1.getModifiers())) {
        Model model = createEmptyModel();
        Element element = class1.getConstructor(String.class).newInstance("x");
        element.setName("test name");
        element.setX(10);
        element.setWidth(10);
        element.setY(10);
        element.setHeight(10);
        model.addElement(element);
        Model deserializedModel = getModelAfterSerializing(model);

        assertEquals("Class " + class1 + " not exported/imported properly", 0,
            comparator.compare(model, deserializedModel));
      }
    }
  }

  @Test
  public void testSetUsedExtensions() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    exporter.removeSbmlExtensions(Arrays.asList(SbmlExtension.values()));
    exporter.addSbmlExtension(SbmlExtension.LAYOUT);

    assertEquals(1, exporter.getSbmlExtensions().size());
    assertEquals(SbmlExtension.LAYOUT, exporter.getSbmlExtensions().iterator().next());

    exporter.removeSbmlExtension(SbmlExtension.LAYOUT);
    assertEquals(0, exporter.getSbmlExtensions().size());
  }

  @Test
  public void testExportWithLayoutExtension() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    exporter.removeSbmlExtensions(Arrays.asList(SbmlExtension.values()));
    exporter.addSbmlExtension(SbmlExtension.LAYOUT);

    Model model = createModelWithReaction();
    String xml = exporter.toXml(model);

    assertTrue(xml.contains("layout:listOfLayouts"));
  }

  @Test
  public void testExportWithoutLayoutExtension() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    exporter.removeSbmlExtensions(Arrays.asList(SbmlExtension.values()));

    Model model = createModelWithReaction();
    String xml = exporter.toXml(model);

    assertFalse(xml.contains("layout:listOfLayouts"));
  }

  @Test
  public void testExportWithRenderExtension() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    exporter.removeSbmlExtensions(Arrays.asList(SbmlExtension.values()));
    exporter.addSbmlExtension(SbmlExtension.LAYOUT);
    exporter.addSbmlExtension(SbmlExtension.RENDER);

    Model model = createModelWithReaction();
    String xml = exporter.toXml(model);

    assertTrue(xml.contains("render:listOfRenderInformation"));
  }

  @Test
  public void testExportWithoutRenderExtension() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    exporter.removeSbmlExtensions(Arrays.asList(SbmlExtension.values()));
    exporter.addSbmlExtension(SbmlExtension.LAYOUT);

    Model model = createModelWithReaction();
    String xml = exporter.toXml(model);

    assertFalse(xml.contains("render:listOfRenderInformation"));
  }

  @Test
  public void testExportWithoutMultiExtensionSupportSpeciesTypes() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    exporter.removeSbmlExtensions(Arrays.asList(SbmlExtension.values()));
    exporter.addSbmlExtension(SbmlExtension.MULTI);

    org.sbml.jsbml.Model sbmlModel = exporter.toSbmlDocument(createModelWithReaction()).getModel();
    MultiModelPlugin multiPlugin = (MultiModelPlugin) sbmlModel.getExtension("multi");
    assertNotNull("Multi plugin is not present in sbml model", multiPlugin);

    assertTrue("Species types are not exported", multiPlugin.getListOfSpeciesTypes().size() > 0);
  }

  @Test
  public void testMultiExtensionSBOTermsForTypes() throws Exception {
    SbmlExporter exporter = new SbmlExporter();
    SBMLDocument doc = new SBMLDocument(3, 2);
    org.sbml.jsbml.Model result = doc.createModel("id");
    MultiModelPlugin multiPlugin = exporter.createSbmlMultiPlugin(result);

    for (MultiSpeciesType speciesType : multiPlugin.getListOfSpeciesTypes()) {
      assertNotNull("SBO term not defined for type " + speciesType.getName(), speciesType.getSBOTermID());
      assertFalse("SBO term not defined for type " + speciesType.getName(), speciesType.getSBOTermID().isEmpty());
    }

  }

  @Test
  public void testExportProteinState() throws Exception {
    Model model = createEmptyModel();
    GenericProtein element = createProtein();
    element.setStructuralState("xxx");
    model.addElement(element);
    Model deserializedModel = getModelAfterSerializing(model);

    assertEquals("Structural state not exported/imported properly", 0, comparator.compare(model, deserializedModel));
  }

  @Test
  public void testExportPositionToCompartment() throws Exception {
    Model model = createEmptyModel();
    GenericProtein element = createProtein();
    element.setPositionToCompartment(PositionToCompartment.INSIDE);
    model.addElement(element);
    Model deserializedModel = getModelAfterSerializing(model);

    assertEquals("Postion to compartment not exported/imported properly", 0,
        comparator.compare(model, deserializedModel));
  }

  @Test
  public void testExportOuterSurfacePositionToCompartment() throws Exception {
    Model model = createEmptyModel();
    GenericProtein element = createProtein();
    element.setPositionToCompartment(PositionToCompartment.OUTER_SURFACE);
    model.addElement(element);
    Model deserializedModel = getModelAfterSerializing(model);

    assertEquals("Postion to compartment not exported/imported properly", 0,
        comparator.compare(model, deserializedModel));
  }

  @Test
  public void testMultiExtensionProteinStateInTypes() throws Exception {
    String structuralState = "xxx";
    Model model = createEmptyModel();
    GenericProtein element = new GenericProtein("id");
    element.setName("test name");
    element.setX(10);
    element.setWidth(10);
    element.setY(10);
    element.setHeight(10);
    element.setStructuralState(structuralState);
    model.addElement(element);
    org.sbml.jsbml.Model sbmlModel = exporter.toSbmlDocument(model).getModel();
    MultiModelPlugin multiPlugin = (MultiModelPlugin) sbmlModel.getExtension("multi");

    boolean structuralStateValueFound = false;
    for (MultiSpeciesType speciesType : multiPlugin.getListOfSpeciesTypes()) {
      for (SpeciesFeatureType featureType : speciesType.getListOfSpeciesFeatureTypes()) {
        for (PossibleSpeciesFeatureValue featureValue : featureType.getListOfPossibleSpeciesFeatureValues()) {
          if (featureValue.getName().equals(structuralState)) {
            structuralStateValueFound = true;
          }
        }
      }
    }
    assertTrue("Structural state not defined in the list of possible values", structuralStateValueFound);

  }

  @Test
  public void testMultiExtensionProteinStateSuplicateInTypes() throws Exception {
    String structuralState = "xxx";
    Model model = createEmptyModel();
    GenericProtein element = new GenericProtein("id");
    element.setName("test name");
    element.setX(10);
    element.setWidth(10);
    element.setY(10);
    element.setHeight(10);
    element.setStructuralState(structuralState);
    model.addElement(element);
    element = new GenericProtein("id2");
    element.setName("test name");
    element.setX(10);
    element.setWidth(10);
    element.setY(10);
    element.setHeight(10);
    element.setStructuralState(structuralState);
    model.addElement(element);
    org.sbml.jsbml.Model sbmlModel = exporter.toSbmlDocument(model).getModel();
    MultiModelPlugin multiPlugin = (MultiModelPlugin) sbmlModel.getExtension("multi");

    for (MultiSpeciesType speciesType : multiPlugin.getListOfSpeciesTypes()) {
      for (SpeciesFeatureType featureType : speciesType.getListOfSpeciesFeatureTypes()) {
        assertTrue(featureType.getListOfPossibleSpeciesFeatureValues().size() <= 2);
      }
    }
  }

  @Test
  public void testMultiExtensionTypeDefinition() throws Exception {
    Model model = createEmptyModel();
    GenericProtein element = createProtein();
    model.addElement(element);
    org.sbml.jsbml.Model sbmlModel = exporter.toSbmlDocument(model).getModel();

    MultiModelPlugin multiPlugin = (MultiModelPlugin) sbmlModel.getExtension("multi");

    org.sbml.jsbml.Species sbmlSpecies = sbmlModel.getSpecies(0);
    MultiSpeciesPlugin speciesExtension = (MultiSpeciesPlugin) sbmlSpecies.getExtension("multi");
    assertNotNull("Multi extension not defined for species", speciesExtension);
    String speciesTypeString = speciesExtension.getSpeciesType();

    MultiSpeciesType speciesType = null;
    for (MultiSpeciesType type : multiPlugin.getListOfSpeciesTypes()) {
      if (type.getId().equals(speciesTypeString)) {
        speciesType = type;
      }
    }
    assertNotNull("Species type is not set in multi extension", speciesType);

  }

  @Test
  public void testMultiExtensionStructuralStateTypeDefinition() throws Exception {
    String structuralState = "xxx";
    Model model = createEmptyModel();
    GenericProtein element = new GenericProtein("id");
    element.setName("test name");
    element.setX(10);
    element.setWidth(10);
    element.setY(10);
    element.setHeight(10);
    element.setStructuralState(structuralState);
    model.addElement(element);
    org.sbml.jsbml.Model sbmlModel = exporter.toSbmlDocument(model).getModel();

    org.sbml.jsbml.Species sbmlSpecies = sbmlModel.getSpecies(0);
    MultiSpeciesPlugin speciesExtension = (MultiSpeciesPlugin) sbmlSpecies.getExtension("multi");
    assertNotNull("Multi extension not defined for species", speciesExtension);
    assertTrue("structural state feature not defined in multi extension",
        speciesExtension.getListOfSpeciesFeatures().size() > 0);
  }

  @Test
  public void testExportResidue() throws Exception {
    Model model = createEmptyModel();
    GenericProtein element = createProtein();
    Residue mr = new Residue("x1");
    mr.setName("217U");
    mr.setState(ModificationState.PHOSPHORYLATED);
    mr.setPosition(new Point2D.Double(10, 11));
    element.addResidue(mr);
    mr = new Residue("Y");
    mr.setName("218");
    mr.setState(ModificationState.PHOSPHORYLATED);
    mr.setPosition(new Point2D.Double(10, 12));
    element.addResidue(mr);
    mr = new Residue("Z");
    mr.setName("219");
    mr.setState(ModificationState.UBIQUITINATED);
    mr.setPosition(new Point2D.Double(10, 13));
    element.addResidue(mr);
    model.addElement(element);
    Model deserializedModel = getModelAfterSerializing(model);

    GenericProtein protein = deserializedModel.getElementByElementId("id");
    assertEquals("Residues weren't exported/imported properly", 3, protein.getModificationResidues().size());

  }

  private GenericProtein createProtein() {
    GenericProtein element = new GenericProtein("id");
    element.setName("test name");
    element.setX(10);
    element.setWidth(10);
    element.setY(10);
    element.setHeight(10);
    return element;
  }

  @Test
  public void testExportNotes() throws Exception {
    Model model = createEmptyModel();
    model.setNotes("XX");
    Model deserializedModel = getModelAfterSerializing(model);

    assertEquals("Notes weren't exported/imported properly", model.getNotes(), deserializedModel.getNotes());

  }

  @Test
  public void testExportImportOfInputOperator() throws Exception {
    Model originalModel = new CellDesignerXmlParser()
        .createModel(new ConverterParams().filename("testFiles/cell_designer_problems/heterodimer_association.xml"));
    Model model = getModelAfterSerializing(originalModel);
    List<Line2D> lines1 = originalModel.getReactions().iterator().next().getLines();
    List<Line2D> lines2 = model.getReactions().iterator().next().getLines();

    ListComparator<Line2D> comparator = new ListComparator<>(new LineComparator(Configuration.EPSILON));
    assertEquals(0, comparator.compare(lines1, lines2));

  }

  @Test
  public void testExportImportOfAdvancedInputOperator() throws Exception {
    Model originalModel = new CellDesignerXmlParser().createModel(new ConverterParams()
        .filename("testFiles/cell_designer_problems/heterodimer_association_with_additional_reactant.xml"));
    Model model = getModelAfterSerializing(originalModel);
    Reaction r1 = originalModel.getReactions().iterator().next();
    Reaction r2 = model.getReactions().iterator().next();
    Line2D line1 = r1.getCenterLine();
    Line2D line2 = r2.getCenterLine();

    assertEquals(0, new LineComparator().compare(line1, line2));

  }

  @Test
  public void testExportImportOfOutputOperator() throws Exception {
    Model originalModel = new CellDesignerXmlParser()
        .createModel(new ConverterParams().filename("testFiles/cell_designer_problems/dissociation.xml"));
    Model model = getModelAfterSerializing(originalModel);
    List<Line2D> lines1 = originalModel.getReactions().iterator().next().getLines();
    List<Line2D> lines2 = model.getReactions().iterator().next().getLines();

    ListComparator<Line2D> comparator = new ListComparator<>(new LineComparator(Configuration.EPSILON));
    assertEquals(0, comparator.compare(lines1, lines2));

  }

  @Test
  public void testExportImportBooleanGate() throws Exception {
    Model originalModel = new CellDesignerXmlParser()
        .createModel(new ConverterParams().filename("testFiles/cell_designer_problems/boolean_logic_gate.xml"));
    Model model = getModelAfterSerializing(originalModel);

    new CellDesignerXmlParser().toXml(model);
    List<Line2D> lines1 = originalModel.getReactions().iterator().next().getLines();
    List<Line2D> lines2 = model.getReactions().iterator().next().getLines();

    ListComparator<Line2D> comparator = new ListComparator<>(new LineComparator(Configuration.EPSILON));
    assertEquals(0, comparator.compare(lines1, lines2));
  }

  @Test
  public void testExportIronMetabolismReaction() throws Exception {
    try {
      Model originalModel = new CellDesignerXmlParser()
          .createModel(new ConverterParams().filename("testFiles/cell_designer_problems/iron_metabolism_reaction.xml"));
      exporter.setProvideDefaults(false);
      String xml = exporter.toXml(originalModel);
      assertNotNull(xml);
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }

  }

}
