package lcsb.mapviewer.web.config;

import lcsb.mapviewer.api.SpringRestApiConfig;
import lcsb.mapviewer.web.security.MvAuthenticationFailureHandler;
import lcsb.mapviewer.web.security.MvAuthenticationSuccessHandler;
import lcsb.mapviewer.web.security.MvUsernamePasswordAuthenticationFilter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.AuthenticationProvider;
import org.springframework.security.config.BeanIds;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.builders.WebSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter;
import org.springframework.security.core.session.SessionRegistry;
import org.springframework.security.web.authentication.LoginUrlAuthenticationEntryPoint;
import org.springframework.security.web.authentication.UsernamePasswordAuthenticationFilter;
import org.springframework.security.web.authentication.session.*;
import org.springframework.security.web.context.HttpSessionSecurityContextRepository;
import org.springframework.security.web.session.ConcurrentSessionFilter;
import org.springframework.security.web.session.SessionManagementFilter;

import java.util.Arrays;

@Configuration
@Import({SpringRestApiConfig.class})
@EnableWebSecurity
public class SpringSecurityConfig extends WebSecurityConfigurerAdapter {

  private SessionRegistry sessionRegistry;

  private AuthenticationProvider authenticationProvider;

  @Autowired
  public SpringSecurityConfig(SessionRegistry sessionRegistry,
                              AuthenticationProvider authenticationProvider) {
    this.sessionRegistry = sessionRegistry;
    this.authenticationProvider = authenticationProvider;
  }

  @Bean
  public SessionAuthenticationStrategy sessionAuthenticationStrategy() {
    ConcurrentSessionControlAuthenticationStrategy strategy1 =
        new ConcurrentSessionControlAuthenticationStrategy(sessionRegistry);
    strategy1.setMaximumSessions(-1);
    strategy1.setExceptionIfMaximumExceeded(true);

    SessionFixationProtectionStrategy strategy2 = new SessionFixationProtectionStrategy();

    RegisterSessionAuthenticationStrategy strategy3 = new RegisterSessionAuthenticationStrategy(sessionRegistry);

    return new CompositeSessionAuthenticationStrategy(Arrays.asList(
      strategy1,
      strategy2,
      strategy3
    ));
  }

  @Override
  public void configure(WebSecurity web) {
    web.ignoring()
        .antMatchers("/resources/css/**")
        .antMatchers("/resources/js/**")
        .antMatchers("/resources/js_tests/**")
        .antMatchers("/resources/images/**")
        .antMatchers("/resources/other/**")
        .antMatchers("/javax.faces.resource/**");
  }

  @Bean
  public HttpSessionSecurityContextRepository httpSessionSecurityContextRepository() {
    return new HttpSessionSecurityContextRepository();
  }

  @Bean
  public MvAuthenticationSuccessHandler authenticationSuccessHandler() {
    MvAuthenticationSuccessHandler handler = new MvAuthenticationSuccessHandler();
    handler.setDefaultTargetUrl("/index.xhtml");
    return handler;
  }

  @Bean
  public MvAuthenticationFailureHandler authenticationFailureHandler() {
    return new MvAuthenticationFailureHandler("/index.xhtml");
  }

  @Bean
  public SessionManagementFilter sessionManagementFilter() {
    SessionManagementFilter filter = new SessionManagementFilter(httpSessionSecurityContextRepository());
    filter.setAuthenticationFailureHandler(authenticationFailureHandler());
    return filter;
  }

  @Bean
  public ConcurrentSessionFilter concurrentSessionFilter() {
    return new ConcurrentSessionFilter(sessionRegistry, "/index.xhtml");
  }

  @Bean(name = BeanIds.AUTHENTICATION_MANAGER)
  @Override
  public AuthenticationManager authenticationManagerBean() throws Exception {
    return super.authenticationManagerBean();
  }

  @Bean
  public UsernamePasswordAuthenticationFilter usernamePasswordAuthenticationFilter() throws Exception {
    MvUsernamePasswordAuthenticationFilter filter = new MvUsernamePasswordAuthenticationFilter();
    filter.setAuthenticationManager(authenticationManagerBean());
    filter.setSessionAuthenticationStrategy(sessionAuthenticationStrategy());
    filter.setAuthenticationSuccessHandler(authenticationSuccessHandler());
    filter.setAuthenticationFailureHandler(authenticationFailureHandler());
    return filter;
  }

  @Bean
  public LoginUrlAuthenticationEntryPoint loginUrlAuthenticationEntryPoint() {
    return new LoginUrlAuthenticationEntryPoint("/index.xhtml");
  }

  @Override
  public void configure(HttpSecurity http) throws Exception {
    http

        .httpBasic()
        .authenticationEntryPoint(loginUrlAuthenticationEntryPoint())

        .and()
        .authenticationProvider(authenticationProvider)

        .addFilterBefore(sessionManagementFilter(), SessionManagementFilter.class)
        .addFilterAt(concurrentSessionFilter(), ConcurrentSessionFilter.class)
        .addFilterAt(usernamePasswordAuthenticationFilter(), UsernamePasswordAuthenticationFilter.class)

        .sessionManagement()
        .sessionAuthenticationStrategy(sessionAuthenticationStrategy())

        .and()
        .authorizeRequests()
        .antMatchers("/login**").permitAll()
        .antMatchers("/index**").permitAll()
        .antMatchers("/export**").permitAll()
        .antMatchers("/galaxy**").permitAll()
        .antMatchers("/map**").permitAll()
        .antMatchers("/map/**").permitAll()
        .antMatchers("/api**").permitAll()
        .antMatchers("/api/**").permitAll()
        .antMatchers("/fonts/**").permitAll()
    		.antMatchers("/plugins/**").permitAll()
        .antMatchers("/**").authenticated()

        .and()
        .headers()
        .frameOptions().disable()

        .and()
        .csrf().disable()

        .logout()
        .logoutUrl("/j_spring_security_logout")
        .logoutSuccessUrl("/index.xhtml")
        .invalidateHttpSession(true);

  }
}
